
=head1 FSDBM_File.pm  (V0.0.1) 

Simple direct-filesystem tied hash interface.

    Usage:
        tie(%hash, 'FSDBM_File', 'in_fact/its_really_a_dir');
    
    Keyname restrictions:
        (to-do - document)

    History:
        BWK  6 May 00 -- Written; Bill Kelly <billk@cts.com>

    Permission to copy, use, modify, sell and distribute this software
    is granted provided this copyright notice appears in all copies.
    Copyright (C) 2000 Bill Kelly / Full Body Groove.  All Rights Reserved.
    This software is released in the hope it will be useful, but please
    understand you use it at your own risk.
    THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
    WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
    OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

=cut


package FSDBM_File;
use strict;
use Fcntl ':flock';  # import LOCK_* constants
use IO::Handle;
use Carp;

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

    $VERSION     = 1.00;

    @ISA         = qw(Exporter);
    @EXPORT      = ( );
    %EXPORT_TAGS = ( );
    @EXPORT_OK   = ( );
}


my $m_dbLockFilename = ".FSDBM_File.lock";
# Undefined keys are set to the following kludge-signature.
# (32 bytes consisting of binary-coded-decimal Pi, with a text identifier
# sandwiched in the middle :)
my $m_undefKludgeVal = "\x31\x41\x59\x26\x53\x58\x97\x93"
                      ."FSDBM_File.undef"
                      ."\x23\x84\x62\x64\x33\x83\x27\x95";

sub whowasi { (caller(1))[3] }

sub TIEHASH {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    
    croak "Usage: @{[&whowasi]} database_root_dir" unless (scalar(@_) == 1);
    my $dbRoot = normalizePath(shift);

    my $self = {
        DBROOT     => $dbRoot,
        DBLOCK_FH  => IO::Handle->new(),
        DBLOCK_REF => 0,  # reference count
    };
    bless ($self, $class);

    $self->_dbCreateOrOpen();

    return $self;
}

sub FETCH {
    my $self = shift;
    my ($key) = @_;

    my $keyPath = $self->_getKeyPath($key);
    my $keyVal = undef;
    my $fh = IO::Handle->new();
    $self->LockDb();
    eval {
        if (-f $keyPath) {
            local($/) = undef;
            open($fh, "$keyPath")
                or croak "Error: @{[&whowasi]} Couldn't open keyfile $keyPath for read";
            $keyVal = <$fh>;
            close($fh);
        }
    };
    my $err = $@;
    $self->UnlockDb();
    undef $fh;
    croak "$err" if ($err);  # rethrow exception if any
    $keyVal = undef if (defined($keyVal)  &&  $keyVal eq $m_undefKludgeVal);
    return $keyVal;
}

sub EXISTS {
    my $self = shift;
    my ($key) = @_;

    my $keyPath = $self->_getKeyPath($key);
    $self->LockDb();  # lock even here to honor atomic transactions in progress by other processes
    my $keyExists = (-f $keyPath);
 # print "EXISTS ($key) ($keyPath) ($keyExists)\n";
    $self->UnlockDb();
    return $keyExists;
}

sub STORE {
    my $self = shift;
    my ($key, $keyVal) = @_;

    $keyVal = $m_undefKludgeVal unless defined($keyVal);
    my $keyPath = $self->_getKeyPath($key);
    createPathToFile($keyPath);  # OK outside of lock
    my $fh = IO::Handle->new();
    $self->LockDb();
    eval {
        open($fh, ">$keyPath")
            or croak "Error: @{[&whowasi]} Couldn't open keyfile $keyPath for write";
        print $fh ($keyVal);
        close($fh);
    };
    my $err = $@;
    $self->UnlockDb();
    undef $fh;
    croak "$err" if ($err);  # rethrow exception if any
    return $keyVal;  # or could return undef(?)
}

sub DELETE {
    my $self = shift;
    my ($key) = @_;

    my $keyPath = $self->_getKeyPath($key);
    $self->LockDb();
    unlink($keyPath) if (-f $keyPath);
    $self->UnlockDb();
    # Untied hash semantics returns value-before-deletion, but the docs
    # indicate tied hashes don't necessarily have to.  Not sure whether
    # we should or not, but for now - we won't.  (To return the value
    # we'd have to read it off disk.)
    return undef;
}

sub CLEAR {
    my $self = shift;

    # Note: Like the example in the perltie docs, we should probably
    # disable the clear by default, and only allow it if the user
    # specifically requests it be enabled.
    # IN FACT, with this operation equating literally to an
    # 'rm -rf $dbRoot' I wouldn't want to chance it at all.
    croak "Error: @{[&whowasi]} Operation not supported (considered too dangerous)";
}

sub FIRSTKEY {
    my $self = shift;

    my $dbRoot = $self->{DBROOT};
    my @keyList = findFiles($dbRoot);
    my $key = shift(@keyList);
    $self->{_FIRSTKEY_keyList} = \@keyList if defined($key);
 # print "FIRSTKEY: (".(defined($key)? $key: "undef").") \[@keyList\]\n";
    return $key;
}

sub NEXTKEY {
    my $self = shift;
  # my ($lastKey) = @_;

    my $keyList = $self->{_FIRSTKEY_keyList};
    my $key = shift(@$keyList);
    $self->{_FIRSTKEY_keyList} = undef unless defined($key);
 # print "NEXTKEY: (".(defined($key)? $key: "undef").") \[@$keyList\]\n";
    return $key;
}

sub DESTROY {
    my $self = shift;

    $self->_dbClose();
}

sub LockDb {
    my $self = shift;
    if (! $self->{DBLOCK_REF}) {
        flock($self->{DBLOCK_FH}, LOCK_EX)
            or croak "Error: @{[&whowasi]} flock() failed obtaining master database lock";
    }
    $self->{DBLOCK_REF}++;
}

sub UnlockDb {
    my $self = shift;
    if ($self->{DBLOCK_REF}) {
        flock($self->{DBLOCK_FH}, LOCK_UN) unless (-- $self->{DBLOCK_REF});
    }
    else {
        carp "Warning: @{[&whowasi]} Unlock with no existing lock(s)";
    }
}

sub _dbCreateOrOpen {
    my $self = shift;
    # Create path to database root if necessary
    my $dbRoot = $self->{DBROOT};
    createPath($dbRoot);
    # Open master database lock file (creating if necessary)
    my $dbLock = "$dbRoot/$m_dbLockFilename";
    open($self->{DBLOCK_FH}, ">>$dbLock")
        or croak "Error: @{[&whowasi]} Couldn't open or create master database lock $dbLock";
}

sub _dbClose {
    my $self = shift;
    # warn if outstanding locks on db
    if ($self->{DBLOCK_REF}) {
        carp "Warning: @{[&whowasi]} outstanding database locks at close: $self->{DBLOCK_REF}";
        $self->{DBLOCK_REF} = 1;  # force next unlock to take
        $self->UnlockDb();
    }
    close($self->{DBLOCK_FH}) if defined($self->{DBLOCK_FH});
}

sub _getKeyPath {
    my $self = shift;
    my ($key) = @_;
    my $dbRoot = $self->{DBROOT};
    return "$dbRoot/$key";
}


=head2 normalizePath (string $path)

Return copy of $path with trailing slash removed, if any.

=cut
sub normalizePath {
    my ($path) = @_;
    $path =~ s{/$}{};  # remove trailing slash, if any
    return $path;
}


=head2 createPath (string $path)

Create every directory component in path, when not already existing.
Die if any component exists, but is not a directory.

=cut
sub createPath {
    my ($path) = @_;
    return if (-d $path);
    my $dir = "";
    while ($path =~ m{(^|/)[^/]+}g) {
        $dir .= $&;
        if (-e $dir) {
            croak "Error: @{[&whowasi]} existing path component is non-directory"
                unless (-d $dir);
        }
        else {
            my $ok = mkdir($dir, 0777);
            croak "Error: @{[&whowasi]} mkdir $dir for path $path failed"
                unless ($ok);
        }
    }
}


=head2 createPathToFile (string $fullPath)

Create every directory component in path, when not already existing.
Die if any directory component exists, but is not a directory.
Final component of path is treated as the filename, and is ignored.

=cut
sub createPathToFile {
    my ($fullPath) = @_;
    createPath($1) if ($fullPath =~ m{^(.*)/.*$});
}



sub findFiles {
    my ($rootPath) = @_;

    $rootPath = normalizePath($rootPath);    
    croak "Error: @{[&whowasi]} given path $rootPath is non-directory"
        unless (-d $rootPath);

    my @fileList = ();
    my $hasSubdirs = ((stat($rootPath))[3] != 2);
    _recursiveFindFiles($rootPath, "", $hasSubdirs, \@fileList);
    return @fileList;
}

sub _recursiveFindFiles {
    my ($fullPath, $relPath, $hasSubdirs, $fileList) = @_;
    
    if (! opendir(DIR, $fullPath)) {
        carp "Warning: @{[&whowasi]} can't read dir $fullPath";
        return;
    }
    my @entries = readdir(DIR);
    closedir(DIR);

    my $entry;
    foreach $entry (@entries) {
        next if ($entry eq '.'  ||  $entry eq '..'  ||
                 $entry eq $m_dbLockFilename);
        my $entryFullPath = "$fullPath/$entry";
        my $entryRelPath = "$relPath$entry";
        if ($hasSubdirs  &&  -d $entryFullPath) {
            my $subHasSubdirs = ((stat(_))[3] != 2);
            _recursiveFindFiles($entryFullPath, "$entryRelPath/", $subHasSubdirs, $fileList);
        }
        else {
            push(@$fileList, $entryRelPath);
        }
    }
}


1;

# eof

