
=head1 FulltextIndexer_Adaptor.pm  (V0.0.4)

Incremetal fulltext-indexing tied hash interface.

    Usage:
        my %ftDatabase;
        my %ftIndexer;
        
        tie(%ftDatabase, AnyDBM_File, ...);
        tie(%ftIndexer, FulltextIndexer_Adaptor,
            \%ftDatabase, 0, "stopwords.txt", \&myWordStemmer);

        $ftIndexer{documentName} = $documentText;  # index document text
        $resultsHash = $ftIndexer{someword};  # return all documents containing someword
        ($docName, $numOccurrances) = each(%$resultsHash);

    Keyname restrictions:
        (to-do - document)


    History:
        BWK  7 May 00 -- Written; Bill Kelly <billk@cts.com>
        BWK  8 May 00 -- Add subdir distribution in wordToKey() (V0.0.2)
        BWK 10 May 00 -- Added public IsStopword() (V0.0.3)
        BWK 13 May 00 -- Added tracking of occurance-count of keywords
                         per document; added support for optional
                         external word-stemmer. (V0.0.4)

    Permission to copy, use, modify, sell and distribute this software
    is granted provided this copyright notice appears in all copies.
    Copyright (C) 2000 Bill Kelly / Full Body Groove.  All Rights Reserved.
    This software is released in the hope it will be useful, but please
    understand you use it at your own risk.
    THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
    WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
    OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

=cut


package FulltextIndexer_Adaptor;
use strict;
use IO::Handle;
use Carp;

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

    $VERSION     = 1.00;

    @ISA         = qw(Exporter);
    @EXPORT      = ( );
    %EXPORT_TAGS = ( );
    @EXPORT_OK   = ( );
}


sub whowasi { (caller(1))[3] }

sub TIEHASH {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    
    croak "Usage: @{[&whowasi]} \\\%storage_hashref use_locking_flag stopwords_file [stemmer_func]\n"
         .'Example: tie(%interfaceHash, FulltextIndexer_Adaptor, \%storageDatabase, 1, "stopwords.txt", \&myStemmer);'
        unless (scalar(@_) == 3  ||  scalar(@_) == 4);  # stemmer optional
    my ($storage, $useLocking, $stopFile, $stemmer) = @_;

    my %stopwords;
    loadStopwordsFile($stopFile, \%stopwords);

    my $self = {
        STORAGE    => $storage,
        LOCKING    => $useLocking,
        STOPWORDS  => \%stopwords,
        STEMMER    => $stemmer,
    };
    bless ($self, $class);
    return $self;
}

sub FETCH {
    my $self = shift;
    my ($wordName) = @_;

    $wordName = lc($wordName);
    my $stemmer = $self->{STEMMER};
    $wordName = &$stemmer($wordName) if defined($stemmer);
    my $key = wordToKey($wordName);
    my $docNames = {};
    $self->LockDb();
    eval {
        $self->_fetchSimpleHash($key, $docNames);
    };
    my $err = $@;
    $self->UnlockDb();
    croak "$err" if ($err);  # rethrow exception if any
    return $docNames;  # values are count of occurrances of corresponding key/words in document
}

sub STORE {
    my $self = shift;
    my ($docName, $fullText) = @_;

    # make literal keyname from $docKey
    # get stopword-stripped unique wordlist from $fullText
    # fetch previous (currently-stored) wordlist from $key
    # remove non-new old words
    # add non-old new words

    my $key = docToKey($docName);
    my %newWords;
    $fullText = "" unless defined($fullText);
    uniqueWordsFromFulltext($fullText, $self->{STOPWORDS},
                            $self->{STEMMER}, \%newWords);

    $self->LockDb();
    eval {
        my %oldWords;
        $self->_fetchSimpleHash($key, \%oldWords);
        my $word;  # !heh
        foreach $word (keys %oldWords) {
            next if defined($newWords{$word});
            $self->_removeWordRef($word, $docName);
        }
        foreach $word (keys %newWords) {
            next if defined($oldWords{$word});
            $self->_addWordRef($word, $docName, $newWords{$word});
        }
        $self->_storeSimpleHash($key, \%newWords);
    };
    my $err = $@;
    $self->UnlockDb();
    croak "$err" if ($err);  # rethrow exception if any
    return 1;  # just so retval isn't 'random'
}

sub EXISTS {
    my $self = shift;
    my ($docName) = @_;

    my $key = docToKey($docName);
    return exists $self->{STORAGE}->{$key};
}

sub DELETE {
    my $self = shift;
    my ($docName) = @_;

    $self->LockDb();
    eval {
        $self->STORE($docName, undef);  # remove this doc's word references
        my $key = docToKey($docName);
        delete $self->{STORAGE}->{$key};
    };
    my $err = $@;
    $self->UnlockDb();
    croak "$err" if ($err);  # rethrow exception if any
    return undef;
}

sub FIRSTKEY {
    my $self = shift;
    return $self->_nextKey();
}

sub NEXTKEY {
    my $self = shift;
  # my ($lastKey) = @_;
    return $self->_nextKey();
}

sub CLEAR {
    my $self = shift;
    croak "Error: @{[&whowasi]} Operation not supported (considered too dangerous)";
}

sub DESTROY {
    my $self = shift;

    # Currently, all allocated resources automatically
    # cleaned up when our hash goes away
}

sub LockDb {
    my $self = shift;
    if ($self->{LOCKING}) {
        tied(%{$self->{STORAGE}})->LockDb();
    }
}

sub UnlockDb {
    my $self = shift;
    if ($self->{LOCKING}) {
        tied(%{$self->{STORAGE}})->UnlockDb();
    }
}

sub IsStopword {
    my $self = shift;
    my ($word) = @_;
    return defined($self->{STOPWORDS}->{$word});
}

sub _nextKey {
    my $self = shift;
    my $stor = $self->{STORAGE};
    my ($key) = each %$stor;
    while (defined($key)  &&  ! isDocKey($key)) {
        ($key) = each %$stor;
    }
    $key = keyToDoc($key) if defined($key);
 # print "_nextKey: (".(defined($key)? $key: "undef").")\n";
    return $key;
}

sub _removeWordRef {
    my $self = shift;
    my ($word, $docName) = @_;

    my $key = wordToKey($word);
    my %wordRefs;
    $self->_fetchSimpleHash($key, \%wordRefs);
    delete $wordRefs{$docName};
    if (scalar keys %wordRefs) {
        $self->_storeSimpleHash($key, \%wordRefs);
    }
    else {
        delete $self->{STORAGE}->{$key};  # word no longer appears in any indexed document
    }
}

sub _addWordRef {
    my $self = shift;
    my ($word, $docName, $numOccurances) = @_;

    my $key = wordToKey($word);
    my %wordRefs;
    $self->_fetchSimpleHash($key, \%wordRefs);
    $wordRefs{$docName} = $numOccurances;
    $self->_storeSimpleHash($key, \%wordRefs);
}

sub _fetchSimpleHash {
    my $self = shift;
    my ($key, $hash) = @_;

    my $stor = $self->{STORAGE};
    my $hashStr = $$stor{$key};
    $hashStr = "" unless defined($hashStr);
    %$hash = split("\012", $hashStr, -1);  # -1 = split even trailing null field(s)
}

sub _storeSimpleHash {
    my $self = shift;
    my ($key, $hash) = @_;

    my $hashStr = join("\012", %$hash);
    my $stor = $self->{STORAGE};
    $$stor{$key} = $hashStr;
}

sub docToKey {
    my ($docName) = @_;
    return ".keys./$docName";
}

sub wordToKey {
    my ($wordName) = @_;
    $wordName = lc($wordName);
    my $sub;  # provide two-letter subdirectories for fewer words in each dir
    if ($wordName =~ /^(.{1,2})/) {
        $sub = $1;
        $sub .= '_' if (length($sub) == 1);
    }
    else {
        $sub = "__";
    }
    return ".words./$sub/$wordName";
}

sub keyToDoc {
    my ($key) = @_;
    $key =~ s{^\.keys\./}{};
    return $key;
}

# sub keyToWord {
#     my ($key) = @_;
#     $key =~ s{^\.words\./../}{};
#     return $key;
# }

sub isDocKey {
    my ($key) = @_;
    return $key =~ m{^\.keys\./};
}

sub uniqueWordsFromFulltext {
    my ($fullText, $stopwords, $stemmer, $keepwords) = @_;

    $fullText = lc($fullText);
    $fullText =~ s/\b([a-z])[+][+]/uc($1)."PP"/gse;  # consider c++, j++ etc. words
    my @words = split(/[^a-zA-Z0-9]+/, $fullText);  # underscore not included
    my $word;
    foreach $word (@words) {
        $word =~ s/([A-Z])PP/lc($1)."++"/ge;  # restore c++, j++, etc.
        next if (length($word) < 2);
        next if defined ($$stopwords{$word});
        $word = &$stemmer($word) if defined($stemmer);
        $$keepwords{$word}++;  # value will be number of occurances of word in text
    }
}

sub loadStopwordsFile {
    my ($stopFile, $stopwords) = @_;

    my $fh = IO::Handle->new();
    open($fh, $stopFile)
        or croak "Error: @{[&whowasi]} Couldn't open stopwords file $stopFile";
    my $line;
    while (defined($line = <$fh>)) {
        chomp $line;
        $line =~ s/#.*$//;  # strip comments
        next unless $line =~ /(\w+)/;
        my $wd = lc($1);
        $$stopwords{$wd} = 1;
    }
    close($fh);
}


1;

# eof

