
=head1 QwikWiki

    History:
        BWK  7 May 00 -- Begun; Bill Kelly <billk@cts.com>
        BWK 22 May 00 -- addRecentChange now deletes only the current
                         author's prior page revision line within
                         today's date - used to match any author. (V0.322)
        BWK 23 May 00 -- HTML header now generated from template in
                         qwcfg; added separate tracking of author IP
                         for save collision detection - some users
                         DNS names change every time they post
                         (proxy1, proxy2, proxy3... etc.) (V0.33)
        BWK 23 May 00 -- External links in markupUrl now given target="_top"
                         to break out of frames. (V0.331)

    Permission to copy, use, modify, sell and distribute this software
    is granted provided this copyright notice appears in all copies.
    Copyright (C) 2000 Bill Kelly / Full Body Groove.  All Rights Reserved.
    This software is released in the hope it will be useful, but please
    understand you use it at your own risk.
    THIS SOFTWARE IS PROVIDED "AS IS", WITHOUT ANY EXPRESS OR IMPLIED
    WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES
    OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.

=cut

package qwiki;
use strict;
use POSIX qw(strftime);
use Socket;  # for gethostbyaddr
use CGI ':standard';
require qwdiff;
require qwcfg;  # QwikWiki config

BEGIN {
    use Exporter   ();
    use vars       qw($VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);

    $VERSION     = 0.331;

    @ISA         = qw(Exporter);
    @EXPORT      = qw(&qwikiInit &viewPage &diffPage
                      &editPage &editCopy &savePage
                      &searchForm &fulltextSearch
                      &validWikiName &validStrictWikiName &detectNonWikiName);
    %EXPORT_TAGS = ( );
    @EXPORT_OK   = qw(&indexWikiPage &simpleHashFromString &simpleHashToString
                      &markupWikiPage &markupWikiPageDiffs
                      &markupWikiLinks &markupUrl &wikiLink
                      &stripWikiMetaCharsInHtml &stripWikiMetaChars);
}

use vars qw( $q );    

sub qwikiInit {
    my ($query) = @_;
    $q = $query;
}

sub docprint {
    local($") = '';
    $qwenv::htmlOut .= "@_";
}

sub viewPage {
    my ($pageName) = @_;

    my $displayName = separateWikiWord($pageName);
    docprint startHtml($displayName);
    my ($pageText, $pageMeta) = fetchPageRev($pageName, 0);
    my $html = defined($pageText)? markupWikiPage($pageText) : "Describe $pageName here";
    my $lastEdited = "";
    if ($$pageMeta{REV} > 1  &&  $$pageMeta{SAVETIME}) {
        my $editDate = getEditDate($$pageMeta{SAVETIME});
        $lastEdited = "(last edited "
                       . a({href=>"$qwcfg::qwikiUrl?diff=$pageName"}, $editDate) . ")";
    }
    _showPage($pageName, $displayName, $html, $lastEdited);
}

sub _showPage {
    my ($pageName, $displayName, $pageHtml, $lastEdited) = @_;

    docprint h1(img({src=>"$qwcfg::qwikiBaseGif"}),
                a({href=>"$qwcfg::qwikiUrl?search=$pageName"}, $displayName));
    docprint $pageHtml, p(hr, a({href=>"$qwcfg::qwikiUrl?edit=$pageName"}, "EditText"),
                                " of this page $lastEdited", br,
                      a({href=>"$qwcfg::qwikiUrl?search=$pageName"}, "FindPage"),
                        " by searching or browsing",
                    # br, wikiLink('NotifyByEmail'), " when page is edited"
                      );
    my $pageRefs = $qwdbm::search{$pageName};
    delete $$pageRefs{$pageName};  # ignore self-references
    if (scalar(keys(%$pageRefs))) {
        my $pageRefText = join(", ", sort keys %$pageRefs);
        markupWikiLinks(\$pageRefText);
        docprint p("This page is mentioned on the following pages: $pageRefText");
    }
}

sub diffPage {
    my ($pageName) = @_;

    my ($oldRev) = $q->param('rev1');
    my ($newRev) = $q->param('rev2');
    my $summarize = defined($q->param('summarize'));
    $oldRev = -1 unless ($oldRev);
    $newRev = 0  unless ($newRev);
    unless ($oldRev =~ /^-?\d+$/) {
        printRequestFailed("Unable to parse '$oldRev' as a number. (rev1=$oldRev)");
        return;
    }
    unless ($newRev =~ /^-?\d+$/) {
        printRequestFailed("Unable to parse '$newRev' as a number. (rev2=$newRev)");
        return;
    }

    tied(%qwdbm::pages)->LockDb();
    my ($oldText, $oldMeta) = fetchPageRev($pageName, $oldRev);
    my ($newText, $newMeta) = fetchPageRev($pageName, $newRev);
    tied(%qwdbm::pages)->UnlockDb();

    my @reason;
    push(@reason, "Revision $oldRev of $pageName not found") unless defined($oldText);
    push(@reason, "Revision $newRev of $pageName not found") unless defined($newText);
    if (scalar @reason) {
        printRequestFailed(join('; ', @reason));
        return;
    }

  # $oldText = "[revision $oldRev not found]\n" unless defined($oldText);
  # $newText = "[revision $newRev not found]\n" unless defined($newText);
    my $html = markupWikiPageDiffs($oldText, $newText, $summarize);
    my $displayName = separateWikiWord($pageName);
    docprint startHtml("$displayName Differences"
                      . (($summarize)? " Summary" : ""));

    my $curTime = time;
    unless ($newRev == 0  &&  $oldRev == -1) {
        docprint timeSinceRevEdit($$oldMeta{REV}, $$oldMeta{SAVETIME}, $curTime,
                                  $$oldMeta{AUTHADDR}), br;
    }
    docprint timeSinceRevEdit($$newMeta{REV}, $$newMeta{SAVETIME}, $curTime,
                              $$newMeta{AUTHADDR}), hr;

    _showPage($pageName, $displayName, $html, "");
}

sub editPage {
    my ($pageName) = @_;

    my ($rev) = $q->param('rev');
    $rev = 0 unless ($rev);
    unless ($rev =~ /^-?\d+$/) {
        printRequestFailed("Unable to parse '$rev' as a number. (rev=$rev)");
        return;
    }

    tied(%qwdbm::pages)->LockDb();
    my ($pageText, $pageMeta) = fetchPageRev($pageName, $rev);
    my $latestMeta = {};
    if ($rev == 0) {
        %$latestMeta = %$pageMeta;
    }
    else {
        $latestMeta = fetchPageMeta(\%qwdbm::pages, $pageName);
    }
    tied(%qwdbm::pages)->UnlockDb();

    if ($$pageMeta{REV} < 1  ||  $$pageMeta{REV} > $$latestMeta{REV}) {
        printRequestFailed("Revision $rev of $pageName not found");
        return;
    }

    my $editingCurrent = ($$pageMeta{REV} == $$latestMeta{REV});
    my $editLabel = ($editingCurrent)? "Edit $pageName" : "Copy of Prior $pageName";
    docprint startHtml($editLabel);
    docprint timeSinceRevEdit($$pageMeta{REV}, $$pageMeta{SAVETIME}, time,
                              $$pageMeta{AUTHADDR}) unless ($editingCurrent);

    my $expectedRev = (exists $qwdbm::pages{$pageName})?
                        ($$latestMeta{REV} + 1) : 1;
    my $allowEditCopy = ($rev == 0);
    $pageText = "Describe $pageName here" unless defined($pageText);
    _editForm($pageName, $pageText, $expectedRev, $editLabel, $allowEditCopy);
}

sub _editForm {
    my ($pageName, $pageText, $expectedRev, $editLabel, $allowEditCopy) = @_;

    docprint $q->startform;
    docprint h1($editLabel, $q->submit(-name => 'FormSave', -value => "Save", -override => 1));
    docprint $q->hidden(-name => 'SavePage', -default => $pageName, -override => 1);
    docprint $q->hidden(-name => 'ExpectedRev', -default => $expectedRev, -override => 1);
  # docprint $q->hidden(-name => 'AuthAddr', -default => $qwenv::clientAddr, -override => 1);
    docprint br, $q->textarea(-name => 'PageText', -default => $pageText,
                              -override => 1, -rows => 20, -columns => 76,
                              -style => 'width:100%', -wrap => 'virtual' );
    docprint br, "<p>Tab-character available here for copying: ";
    docprint $q->textfield(-name => 'TabCharField',
                           -default => "\t",
                           -override => 1,
                           -size => 8,
                           -maxlength => 1);
    docprint br, br;
    docprint wikiLink('GoodStyle'), " tips for editing.", br;
    docprint a({href=>"$qwcfg::qwikiUrl?copy=$pageName"}, "EditCopy"),
             " from previous author.", br
        if ($allowEditCopy);
    docprint "</p>";
    docprint $q->endform;
}

sub editCopy {
    my ($pageName) = @_;
    $q->param(-name => 'rev', -value => '-1') unless ($q->param('rev'));
    editPage($pageName);
}

sub savePage {
    my ($pageName) = @_;
 
    my ($editedText) = $q->param('PageText');
    my ($expectedRev) = $q->param('ExpectedRev');
    my $newAuthIP = $qwenv::clientAddr;
 # $newAuthIP = $1 if ($editedText =~ /authaddr=([\w.-]+)/s);  # %%DEBUG
 # print "<tt>savePage xrev($expectedRev) auth($newAuthIP)</tt><br>";  # %%DEBUG

#while ($editedText =~ /[^\011\012\015\040-\176]/gs) {  # %%DEBUG:
#   print STDOUT qq{<p><font color="FF0000">},
#                sprintf("(DEBUG) Warning: Binary character on save: [%s] (0x%02x)", $&, ord($&)),
#                qq{</font></p>\n};
#}

    $editedText = "" unless($editedText);
    $editedText =~ s/[\015][\012]/\n/gs;
    $editedText =~ s/[\015\012]/\n/gs;
    if ($editedText =~ /^\s*$/s) {
        docprint startHtml("$pageName Not Saved");
        docprint h1(qq{<font color="FF7700">Alert - Page Not Saved</font>});
        docprint p("The page data received was entirely blank. ",
                   "As a precaution, the blank page was not saved.", br, br,
                   em("Tip: In order to save a completely blank page, ",
                      "delete the text and enter just a single back-quote ",
                      "(`) character, then save."));
        return;
    }

    if ($qwcfg::enableStoryMarkup) {
        my $errBuf = "";
        _interpretStoryMarkup($editedText, \$errBuf);
        if ($errBuf) {
            docprint startHtml("$pageName Not Saved");
            docprint h1(qq{<font color="FF7700">Alert - Story Markup Trouble</font> <small>($pageName Not Saved)</small>});
            docprint p("The following problems were encountered with the Story Markup ",
                       "on $pageName:");
            docprint "<ul>", p($errBuf), "</ul>";
            docprint p("Please use the browser back button to return to the ",
                       "edit form and make any corrections.  Thank you.");
            return;
        }
    }

    # Name lookup may take some time - perform before database lock
    my $newAuthAddr = getHostName($newAuthIP);
 
    tied(%qwdbm::pages)->LockDb();
    tied(%qwdbm::revs)->LockDb();
    my $isNewPage = ! (exists $qwdbm::pages{$pageName});
    my $latestMeta = fetchPageMeta(\%qwdbm::pages, $pageName);
    my $latestRev = $$latestMeta{REV};

    # Allow same author to save older edits over newer revs (which will
    # happen if author has used 'back' button in browser to re-save a page.)
    my $collisionMeta = undef;
    unless ($isNewPage) {
        my $rev;
        for ($rev = $expectedRev;  $rev <= $latestRev  &&  ! defined($collisionMeta);  $rev++) {
            my $revMeta = ($rev == $latestRev) ?
                    $latestMeta :
                    fetchPageMeta(\%qwdbm::revs, "$pageName/$rev");
            $collisionMeta = $revMeta unless ($$revMeta{AUTHIP} eq $newAuthIP  ||
                                              $$revMeta{AUTHADDR} eq $newAuthAddr);
     # print "<tt>Checking rev $rev auth $$revMeta{AUTHIP} vs new_auth $newAuthIP: collision=". (defined($collisionMeta)? "yes":"no") ." (latestRev=$latestRev)</tt><br>";
        }
    }

    my $curTime = time;
    if (defined($collisionMeta)) {
        my $origRev = $expectedRev - 1;
        handleSaveCollision($pageName, $editedText, $origRev, $newAuthAddr, $curTime,
                            $collisionMeta, $latestMeta);
        tied(%qwdbm::revs)->UnlockDb();
        tied(%qwdbm::pages)->UnlockDb();
    }
    else {
        # %%BWK 15 May 00 -- strange behavior with tied hash:
        #   The following expression: scalar(keys(%{$qwdbm::search{RandomPage}}))
        #   errors with: Can't use an undefined value as a HASH reference
        #   However, $qwdbm::search ALWAYS returns a real HASH ref, so there
        #   shouldn't be an undefined value involved.  The identical expression
        #   works fine with an untied hash.
        my $pageRefs;
        my $blankAndUnreferenced =
            (($editedText =~ /^\s*[`]\s*$/s)  &&
             defined($pageRefs = $qwdbm::search{$pageName})  &&
             scalar(keys(%$pageRefs)) == 0);
        my $updateType = ($blankAndUnreferenced)? " (removed)" :
                            (($isNewPage)? " (new)" : "");
        my $saveText = ($blankAndUnreferenced)? undef : $editedText;
        my $overwriteOK = $isNewPage;
        savePageRev($pageName, $saveText, $newAuthIP, $newAuthAddr,
                    $curTime, $latestMeta, $overwriteOK);
        addRecentChange($pageName, $updateType, $newAuthAddr, $curTime);
        # Atomic transaction for page+search update
        tied(%qwdbm::search)->LockDb();
        tied(%qwdbm::revs)->UnlockDb();
        tied(%qwdbm::pages)->UnlockDb();
        # Index page text for searching
        indexWikiPage($pageName, $editedText) unless ($pageName eq $qwcfg::RecentChanges);
        tied(%qwdbm::search)->UnlockDb();
        docprint startHtml("Thanks For Editing $pageName");
        docprint "<br><p>Thanks for editing " . wikiLink($pageName) . ". ";
        if ($blankAndUnreferenced) {
            docprint br, br, "Since the text of the saved page was blank, and ",
                     "$pageName was not currently linked-to from any other page, ",
                     "$pageName has removed (hidden) from the database. ",
                     "The prior revision of the page, if any, is still accessible using ",
                     a({href=>"$qwcfg::qwikiUrl?copy=$pageName"}, "EditCopy") . ".", br
                     br, hr, br,
                     a({href=>"$qwcfg::qwikiUrl?search=$pageName"}, "FindPage"),
                       " by searching or browsing", br;
        }
        else {
            docprint "Be sure to <b>reload</b> your old pages.";
        }
        docprint "</p>";
    }
}

# Passing latestMeta into handleSaveCollision is just an (ugly) optimization
# (savePage already had had to fetch it... so not re-fetching it here)
sub handleSaveCollision {
    my ($pageName, $editedText, $origRev, $newAuthAddr, $curTime, $collisionMeta, $latestMeta) = @_;
    docprint startHtml("$pageName Save Collision");
    docprint h1(qq{<font color="FF7700">Alert - Save Collision</font> <small>($pageName Not Saved)</small>});
    my $timeAgo = getDiffTime($$collisionMeta{SAVETIME}, $curTime);
    docprint p("While you were editing, $pageName was saved by <i>$$collisionMeta{AUTHADDR}</i> "
             . "$timeAgo ago.  To keep your changes, you'll need to copy them into "
             . "the &quot;Newer $pageName&quot; edit form below.", br, br, hr);

    my ($origText, $origMeta) = fetchPageRev($pageName, $origRev);
    docprint $q->startform;
    docprint p("This is your edited page, followed by a summary of the "
             . "changes you made.");
 #  docprint h2("<tt><b><u>Your Edit:</u></b></tt>");
    docprint br, $q->textarea(-name => 'EditedText', -default => $editedText,
                              -override => 1, -rows => 8, -columns => 76,
                              -style => 'width:100%', -wrap => 'virtual' );

    my $summarize = 1;
    my $html = markupWikiPageDiffs($origText, $editedText, $summarize);
    docprint h2("<tt><b><u>Change Summary:</u></b></tt>");
    docprint $html;
    docprint $q->endform, hr;

    docprint p("Copy or merge your changes into the newer page here.");

  #  docprint timeSinceRevEdit($$collisionMeta{REV}, $$collisionMeta{SAVETIME}, $curTime,
  #                            $$collisionMeta{AUTHADDR});

    my $latestText = $qwdbm::pages{$pageName};
    $latestText = "" unless ($latestText);
    my $expectedRev = $$latestMeta{REV} + 1;
    my $editLabel = "Edit Newer $pageName";
    my $allowEditCopy = 0;
    _editForm($pageName, $latestText, $expectedRev, $editLabel, $allowEditCopy);
}

sub savePageRev {
    my ($pageName, $pageText, $newAuthIP, $newAuthAddr, $curTime, $oldMeta, $overwrite) = @_;
    my %newMeta = %$oldMeta;
    my $lastRev = $$oldMeta{REV};
    unless ($overwrite) {
        # Copy existing rev to revs db
        $qwdbm::revs{"$pageName/$lastRev"} = $qwdbm::pages{$pageName};
        storePageMeta(\%qwdbm::revs, "$pageName/$lastRev", $oldMeta);
        $newMeta{REV} = ($lastRev + 1);
    }
    $newMeta{AUTHIP}   = $newAuthIP;
    $newMeta{AUTHADDR} = $newAuthAddr;
    $newMeta{SAVETIME} = $curTime;
    if (defined($pageText)) {
        $qwdbm::pages{$pageName} = $pageText;
    }
    else {
        delete $qwdbm::pages{$pageName};
    }
    storePageMeta(\%qwdbm::pages, $pageName, \%newMeta);
}

sub fetchPageRev {
    my ($pageName, $wantRev) = @_;

    tied(%qwdbm::pages)->LockDb();
    my $pageMeta = fetchPageMeta(\%qwdbm::pages, $pageName);
    my $curRev = $$pageMeta{REV};
    my $getRev = ($wantRev < 1)? ($curRev + $wantRev) : $wantRev;
    my $pageText;
    if ($getRev == $curRev) {
        $pageText = $qwdbm::pages{$pageName};
        # already have current meta
    }
    else {
        $pageText = $qwdbm::revs{"$pageName/$getRev"};
        $pageMeta = fetchPageMeta(\%qwdbm::revs, "$pageName/$getRev");
        unless ($$pageMeta{REV} == $getRev) {
            main::logmsg("fetchPageRev: meta rev " . $$pageMeta{REV}
                        ." not equal keyname rev $getRev for page $pageName");
            $$pageMeta{REV} = $getRev;
        }
    }
    tied(%qwdbm::pages)->UnlockDb();
    return ($pageText, $pageMeta);
}

sub fetchPageMeta {
    my ($db, $pageName) = @_;
    my $pageMeta = simpleHashFromString($$db{"$pageName.meta"});
    $$pageMeta{REV} = 1 unless defined($$pageMeta{REV});
    $$pageMeta{AUTHIP}   = "unknown" unless defined($$pageMeta{AUTHIP});
    $$pageMeta{AUTHADDR} = "unknown" unless defined($$pageMeta{AUTHADDR});
    $$pageMeta{SAVETIME} = 0 unless defined($$pageMeta{SAVETIME});
    return $pageMeta;
}

sub storePageMeta {
    my ($db, $pageName, $pageMeta) = @_;
    $$db{"$pageName.meta"} = simpleHashToString($pageMeta);
}

sub addRecentChange {
    my ($editedPageName, $updateType, $editAuthAddr, $saveTime) = @_;

    my ($pageText, $pageMeta) = fetchPageRev($qwcfg::RecentChanges, 0);

    # Add today's date if rollover since last RecentChanges update
    # ($sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst) 
    my @prevTime = localtime $$pageMeta{SAVETIME};
    my @curTime = localtime $saveTime;
    my $newDay = ($curTime[5] != $prevTime[5]  ||
                  $curTime[7] != $prevTime[7]);  # check year(5) and yday(7)
    my $curDate = getEditDate($saveTime);
    $pageText =~ s/^(.*?)\s*$/$1\n/s;  # strip extra trailing whitespace
    my $oldUpdateType;
    if ($newDay) {
        $pageText .= "\n$curDate\n\n";
    }
    else {
        # Remove this author's prior entry for current page within current day, if any
        $pageText =~ s/^(.*\n$curDate\n(?:.*\n)?)\t[*]\s+$editedPageName(\s+[(]new[)])?[^\n]*$editAuthAddr.*?(?:\n|$)/$1/s;
        $oldUpdateType = $2;  # carry-forward today's new-page comment, if any <big kludgy grin>
    }

    # Append new change
    $updateType = $oldUpdateType unless ($updateType  ||  ! defined($oldUpdateType));
    $pageText .= "\t* $editedPageName$updateType . . . . . . $editAuthAddr\n";

    # Preserve last real authAddr of RecentChanges page
    my $overwriteOK = 1;
    savePageRev($qwcfg::RecentChanges, $pageText,
                $$pageMeta{AUTHIP}, $$pageMeta{AUTHADDR},
                $saveTime, $pageMeta, $overwriteOK);
}

sub searchForm {
    my ($pageName) = @_;

    docprint startHtml("Fulltext Search");
    docprint h1("Fulltext Search");
    docprint $q->startform;
    docprint br, p("Enter words and PageNames, or leave blank to list all WikiPages");
    _printQueryInterface($pageName);
    docprint $q->endform;
}

sub fulltextSearch {
    my ($searchQuery) = @_;

    docprint startHtml("Search Results");
    docprint h1("Search Results");

    if ($searchQuery =~ /^\s*$/s) {
      # docprint p(b("Unable to perform search because the query string was empty."));
        my @allPages = sort keys %qwdbm::pages;
        docprint p(b("All Wiki pages (alphabetically)"));
        my $page;
        foreach $page (@allPages) {
            next unless validWikiName($page);   # weed out metadata files
            next unless exists($qwdbm::pages{$page});
            docprint wikiLink($page), br;
        }
    }
    else {
        my $pageHits = performSearch($searchQuery);
        if (scalar keys(%$pageHits)) {
            docprint p(i("To-do: Contextual results display requested, but not implemented yet."
                        ."defaulting to standard search results."))
                if (defined($q->param('ShowContext')) && $q->param('ShowContext') eq 'ON');

            docprint p(b("One or more search terms found on the following pages:"));
            
            my %hitRanks;
            my ($page, $termHits);
            while (($page, $termHits) = each %$pageHits) {
                $hitRanks{$termHits} = [] unless defined($hitRanks{$termHits});
                push(@{$hitRanks{$termHits}}, $page);
            }

            my @rankLevels = sort {$b <=> $a} keys %hitRanks;
            my $rankLevel;
            foreach $rankLevel (@rankLevels) {
                my @pageListAtRank = sort @{$hitRanks{$rankLevel}};
                foreach $page (@pageListAtRank) {
                    docprint "$rankLevel. ", wikiLink($page), br;
                }
            }
        }
        else {
            docprint p("None of the search terms was found in the database.",
                       i("To-do: display any terms that were stopwords.")), br;
        }
    }

    docprint hr;
    docprint $q->startform;
    docprint br, p("Another search?");
    _printQueryInterface($searchQuery);
    docprint $q->endform;
}

sub _printQueryInterface {
    my ($defaultQuery) = @_;

    docprint $q->textfield(-name => 'FulltextSearch',
                           -default => $defaultQuery,
                           -override => 1,
                           -size => 50,
                           -maxlength => 127);
    docprint $q->submit(-name => 'FormSubmit', -value => "Search", -override => 1);
  # docprint br;
  # docprint $q->checkbox(-name => 'ShowContext',
  #                     # -checked => 'checked',
  #                       -value => 'ON',
  #                       -label => 'Show surrounding context with search results (takes longer)'
  #                       -override => 1);
}

sub performSearch {
    my ($searchQuery) = @_;
    
    my @terms = split('\s+', $searchQuery);
    my %uniqueTerms;
    my $term;
    foreach $term (@terms) {
        $uniqueTerms{$term} = 1;
    }
    @terms = keys(%uniqueTerms);

 # print pre("performSearch: terms(@terms)");  # %%DEBUG
    
    my %pageHits;
    foreach $term (@terms) {
        my $pageRefs = $qwdbm::search{$term};
        my $page;
        foreach $page (keys %$pageRefs) {
            $pageHits{$page}++;
 # print pre("performSearch: term($term) hit page($page)");  # %%DEBUG
        }
    }

 # print pre("performSearch: pageHits(", join(',', %pageHits), ")");  # %%DEBUG

    return \%pageHits;
}

sub detectNonWikiName {
    my ($pageName) = @_;
    return undef if validWikiName($pageName);

    printRequestFailed("Unable to parse '$pageName' as a WikiName.");
    return 1;
}

sub printRequestFailed {
    my ($reason) = @_;

    docprint startHtml("QwikWiki V$qwiki::VERSION");
    docprint h2("The Wiki Server Can't Process Your Request");
    docprint p("<tt>&nbsp;&nbsp;&nbsp;&nbsp;$reason</tt>", br, br,
               "We are sorry for any inconvenience.", br, br, br, hr, br
               a({href=>"$qwcfg::qwikiUrl?search"}, "FindPage"),
                       " by searching or browsing");
}

sub startHtml {
    my ($title) = @_;
    my $head = $qwcfg::htmlHeader;
    $head =~ s/%%TITLE%%/$title/gs;
    return $head;
}

sub markupWikiPage {
    my ($plain) = @_;

    $plain =~ s/</&lt;/gs;
    $plain =~ s/>/&gt;/gs;
    $plain = _interpretStoryMarkup($plain) if ($qwcfg::enableStoryMarkup);
    return _markupWikiPage($plain);
}

sub markupWikiPageDiffs {
    my ($old, $new, $summarize) = @_;

    $old =~ s/</&lt;/gs;
    $old =~ s/>/&gt;/gs;
    $new =~ s/</&lt;/gs;
    $new =~ s/>/&gt;/gs;
    my $diff = qwdiff::hilightPageDiffs($old, $new);
    $diff = summarizeDiffs($diff) if ($summarize);
    return _markupWikiPage($diff);
}

sub summarizeDiffs {
    my ($diff) = @_;
    
    $diff .= "\n\n";  # sentinel guarantees flush accum
    my $summary = "";
    my $accum = "";
    my $from;
    my $lno = 0;
    while ($diff =~ /.*?(?:\n|$)/sg) {
        $lno++;
        my $line = $&;
        if ($line =~ /<!--qwdiff-->/s) {
            $from = $lno unless ($from);
            $accum .= $line;
        }
        elsif ($from) {
            my $to = ($lno - 1);
            $summary .= "\n\n<tt><b><u>"
                       . (($to > $from)? "lines $from - $to" : "line $from")
                       . "</u></b></tt><br>\n";
            $summary .= $accum;
            $accum = "";
            undef $from;
        }
    }
    return $summary;
}

sub _markupWikiPage {
    my ($plain) = @_;

    my %listTypeOpen =  ( '*' => "<ul>",  '1.' => "<ol>" );
    my %listTypeClose = ( '*' => "</ul>", '1.' => "</ol>" );
    my %paraTypeOpen =  ( 'p' => "<p>",  'pre' => "<pre>" );
    my %paraTypeClose = ( 'p' => "</p>", 'pre' => "</pre>" );
    my $GT = '(?:<[^>]*>)*';  # "gobble tags"

    my $line = "";
    my $linePlain = "";
    my $html = "";
    my $prevTextInterrupted = 0;
    my $blankCnt = 0;
    my @blankBuf = ();
    my $blankHasDiff = 0;
    my $inPara = "";
    my @listLvl = ();

    my $_isBlank = sub {
        return $linePlain =~ /^\s*$/s;
    };
    my $_isDivider = sub {
        return $linePlain =~ /^----/;
    };
    my $_isMono = sub {
        return $linePlain =~ /^[ ]+/;
    };
    my $_isItem = sub {
        return $linePlain =~ /^(\t+)([*]|1\.)/s;
    };
    my $_closePara = sub {
        return unless ($inPara);
        $prevTextInterrupted = 1;
        $html .= $paraTypeClose{$inPara}."\n";
        $inPara = "";
    };
    my $_openPara = sub {
        my ($pType) = @_;
        return "" if ($inPara eq "$pType");
        &$_closePara;
        $prevTextInterrupted = 1;
        $html .= $paraTypeOpen{$pType}."\n";
        return $inPara = $pType;
    };
    my $_goListLevel = sub {
        my ($tabs, $listType) = @_;
        my $itemLvl = length($tabs);
 # print STDERR ("_goListLevel: ($itemLvl) (", defined($listType)? $listType: "undef", ") (@listLvl)\n");
        while (scalar(@listLvl) > $itemLvl  ||
               ($itemLvl > 0 && (scalar(@listLvl) == $itemLvl) &&
                ($listLvl[$itemLvl-1] ne $listType))) {
            $prevTextInterrupted = 1;
            $html .= $listTypeClose{ pop(@listLvl) };
        }
        while (scalar(@listLvl) < $itemLvl) {
            push(@listLvl, $listType);
            $prevTextInterrupted = 1;
            $html .= $listTypeOpen{ $listType };
        }
    };

    my $prevWasText = 0;
    while ($plain =~ /.*?(\n|$)/gs) {
        $linePlain = $line = $&;
        $linePlain =~ s/($GT)//gso;

        $line =~ s{'''''(.*?)'''''}{<b><i>$1</i></b>}gs;
        $line =~ s{'''(.*?)'''}{<b>$1</b>}gs;
        $line =~ s{''(.*?)''}{<i>$1</i>}gs;
 
        if (&$_isBlank) {
            $blankCnt++;
            if ($line =~ /<!--qwdiff-->/s) {
                my $sp = ($prevWasText)? "" : "&nbsp;";
                $line =~ s/<!--qwdiff_end-->.*?$/$sp<br>$&/s;
                $blankHasDiff = 1;
                push(@blankBuf, $line);
            }
            else {
                push(@blankBuf, "<br>\n");
            }
            $line = "";
            $prevWasText = 0;
        }
        else {
            my $lineWasText = 0;
            if (&$_isItem) {
                &$_closePara;
                $prevTextInterrupted = 1;
                $blankCnt -= 1;  # added because allowing linefeed seemed nicer here in EditText
              # $line =~ s/^(\t+)([*]|1\.)/&$_goListLevel($1, $2); "<li>"/e;
                $line =~ s/^($GT)(\t+)([*]|1\.)/&$_goListLevel($2, $3); "$1<li>"/eo;
            }
            else {
                &$_goListLevel("");
                if (&$_isDivider) {
                    $blankCnt++ unless ($prevTextInterrupted);
                    $prevTextInterrupted = 1;  # (hr has implicit br)
                 #   my $hr = ($line =~ /<!--qwdiff-->/s)?
                 #               "&nbsp;<hr>&nbsp;\n" : "<hr>\n";
                    $line =~ s/-{4,}/<hr>/s;
                    my $tags = "";
                    while ($line =~ /($GT)/gso) {
                        $tags .= $1;
                    }
                    $line = $tags;
                }
                else {
                    my $newPara = &$_openPara( (&$_isMono)? 'pre': 'p');
                    $blankCnt-- if ($newPara);  # new para has implicit br
                    $lineWasText = 1;
                }
            }
            my $numBr = (($blankCnt < 1) || $prevTextInterrupted)?
                                                $blankCnt : ($blankCnt + 1);
            if ($blankHasDiff) {
                $html .=  join('', @blankBuf);
            }
            else {
                $html .= ("<br>\n" x $numBr) if ($numBr > 0);
            }
            $blankCnt = 0;
            @blankBuf = ();
            $blankHasDiff = 0;
            $prevTextInterrupted = 0 if ($lineWasText);
            $prevWasText = $lineWasText;
        }

        markupWikiLinks(\$line);
        stripWikiMetaCharsInHtml(\$line);
        $html .= $line;
    }
    &$_goListLevel("");
    &$_closePara;
    $html .= ("<br>\n" x ($blankCnt-1)) if (($blankCnt-1) > 0);

    return $html;
}

sub _interpretStoryMarkup {
    my ($text, $errBuf) = @_;
    my $_dummyBuf = "";
    $errBuf = \$_dummyBuf unless defined($errBuf);

    return $text unless ($text =~ /((^|\n)``)|(``$)/s);  # no story markup on page?

    $qwenv::random25 = rand unless defined($qwenv::random25);
    $qwenv::random50 = rand unless defined($qwenv::random50);
    $qwenv::random75 = rand unless defined($qwenv::random75);

    my $out = "";
    my $inBlock = 0;
    my $thisCaseSucceeded = 0;
    my $priorCaseSucceeded = 0;
    my $line = "";
    
    my %opExe = (
        'CameFrom'      => sub { return   scalar(grep {$_ eq $qwenv::referrerPage} @_) },
        'NotComingFrom' => sub { return ! scalar(grep {$_ eq $qwenv::referrerPage} @_) },
        'OtherWise'     => sub { return 1 },
        'TwentyFivePercentChance'  => sub { return $qwenv::random25 < 0.25 },
        'FiftyPercentChance'       => sub { return $qwenv::random50 < 0.50 },
        'SeventyFivePercentChance' => sub { return $qwenv::random75 < 0.75 }
    );
    my %needArgs = (
        'CameFrom'      => 1,
        'NotComingFrom' => 1
    );
    my %goodBlockStart = (
        'CameFrom'                  => 1,
        'NotComingFrom'             => 1,
        'TwentyFivePercentChance'   => 1,
        'FiftyPercentChance'        => 1,
        'SeventyFivePercentChance'  => 1
    );
    my %opsNotPreceeding = (
        'CameFrom'                  => { 'OtherWise' => 1, },
        'NotComingFrom'             => { 'OtherWise' => 1, },
        'OtherWise'                 => { 'OtherWise' => 1, },
        'TwentyFivePercentChance'   => { 'OtherWise' => 1, 'TwentyFivePercentChance'  => 1 },
        'FiftyPercentChance'        => { 'OtherWise' => 1, 'FiftyPercentChance'       => 1 },
        'SeventyFivePercentChance'  => { 'OtherWise' => 1, 'SeventyFivePercentChance' => 1 }
    );
    my %opSynonyms = (
        'ComingFrom'                => 'CameFrom'
    );
    my %opsThisBlock = ();

    my $_reportError = sub {
        my ($err) = @_;
        my $errLine = $line;
        chomp ($errLine);
 # print "$err, near:<br>\n$errLine<br>\n";
        $$errBuf .= "<tt><b>$err, near:</b></tt><br>\n$errLine<br><br>\n";
        $out .= qq{<br><font color="FF7700"><big><b>Story Markup Problem</b></big></font><br>\n}
              . qq{<tt><b>$err, near:</b></tt><br>\n}
              . qq{<font style="color:black;background-color:#ffff66">`$errLine</font>\n\n};
    };
    my $_reportErrorBadBlockStart = sub {
        my ($op) = @_;
        &$_reportError("Tag '$op' can't be the first statement of a block");
    };
    my $_reportErrorBadOp = sub {
        my ($op) = @_;
        &$_reportError("Unrecognized Story Markup Tag '$op'");
    };
    my $_reportErrorMisplacedOp = sub {
        my ($op, $badPreceeding) = @_;
        my %bad = map { $_ => 1 } @$badPreceeding;
        &$_reportError("Tag '$op' will never be chosen because it "
                      . (($bad{$op})? "has already appeared earlier in this block" :
                                      "can't follow tags " . join(",", sort keys %bad)
                                    . " in a block"));
    };
    my $_reportErrorNeedArgs = sub {
        my ($op) = @_;
        &$_reportError("Tag '$op' needs one or more PageNames to check");
    };
    my $_reportErrorMisplacedEndBlock = sub {
        &$_reportError("Tag EndBlock can't appear as the beginning of a block");
    };

    my $_blockStart = sub {
        my ($op) = @_;
        $op = $opSynonyms{$op} if defined($opSynonyms{$op});
        &$_reportErrorBadBlockStart($op) unless ($goodBlockStart{$op});
        $inBlock = 1;
        $thisCaseSucceeded = 0;
        $priorCaseSucceeded = 0;
        %opsThisBlock = ( );
    };

    my $_blockEnd = sub {
        $inBlock = 0;
        $priorCaseSucceeded = 0;
        $thisCaseSucceeded = 0;
    };

    my $_execOp = sub {
        my ($op, $param) = @_;
        $op = $opSynonyms{$op} if defined($opSynonyms{$op});
        unless (defined($opExe{$op})) {
            &$_reportErrorBadOp($op);
            return;
        }
        my @badPreceeding = grep { $opsNotPreceeding{$op}->{$_} } keys %opsThisBlock;
        &$_reportErrorMisplacedOp($op, \@badPreceeding) if (scalar(@badPreceeding));
        $opsThisBlock{$op}++;
        
        my @args = split(/[^A-Za-z]/, $param);
        if ($needArgs{$op}  &&  ! scalar(@args)) {
            &$_reportErrorNeedArgs($op);
        }
        else {
            unless ($priorCaseSucceeded) {
                my $func = $opExe{$op};
                if (defined($func)) {
                    $thisCaseSucceeded = &$func(@args);
                }
            }
        }
    };

    my $prevWasBlank = 0;
    while ($text =~ /.*?(\n|$)/gs) {
        $line = $&;

 # print "ISM: top inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";

        if ($line =~ /^``/) {
 # print "ISM: iscmd inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";
            if (! $inBlock) {
                if ($line =~ /^``(EndBlock)?(\s|$)/s) {
                    &$_reportErrorMisplacedEndBlock();
                }
                elsif ($line =~ /^``([A-Za-z]+)\s*(.*?)\s*$/s) {
                    my ($op, $param) = ($1, $2);
 # print "ISM: stblk op($op) param($param) inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";
                    &$_blockStart($op);
                    &$_execOp($op, $param);
                }
                else {
                    &$_reportError("Unrecognized Story Markup Tag");
                }
            }
            else {
                $priorCaseSucceeded |= $thisCaseSucceeded;
                $thisCaseSucceeded = 0;
            
                if ($line =~ /^``(EndBlock)?(\s|$)/s) {
 # print "ISM: end_block\n<br>";
                    &$_blockEnd();
                }
                elsif ($line =~ /^``OtherWise\s*$/s) {
 # print "ISM: other op(OtherWise) param() inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";
                    &$_execOp('OtherWise', "");
                }
                elsif ($line =~ /^``OtherWise\s+([A-Za-z]+)\s*(.*?)\s*$/s) {
                    my ($op, $param) = ($1, $2);
 # print "ISM: otblk op($op) param($param) inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";
                    &$_execOp($op, $param);
                }
                else {
                    &$_reportError( ($line =~ /^``OtherWise/)?
                                    "Unrecognized Story Markup Tag" :
                                    "Story Markup Tag continuing a block must begin with the 'OtherWise' tag");
                }
            }
        }
        else {
 # print "ISM: out inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";
            $out .= $line unless ( ($inBlock &&  ! $thisCaseSucceeded) ||
                                   (! $inBlock  &&  $prevWasBlank  &&
                                     ($line =~ /^\s*$/s) ) );
        }
        $prevWasBlank = ($line =~ /^\s*$/s);
    }

    return $out;
}




#   sub _interpretStoryMarkup {
#       my ($text, $errBuf) = @_;
#       my $_dummyBuf = "";
#       $errBuf = \$_dummyBuf unless defined($errBuf);
#   
#       $qwenv::random = rand unless defined($qwenv::random);
#   
#       my $out = "";
#       my $inBlock = 0;
#       my $thisCaseSucceeded = 0;
#       my $priorCaseSucceeded = 0;
#       my $line = "";
#       
#       my %opExe = (
#           'CameFrom'      => sub { return   scalar(grep {$_ eq $qwenv::referrerPage} @_) },
#           'NotComingFrom' => sub { return ! scalar(grep {$_ eq $qwenv::referrerPage} @_) },
#           'OtherWise'     => sub { return 1 },
#           'ReMark'        => sub { return 0 },
#           'TwentyFivePercentChance'  => sub { return $qwenv::random < 0.25 },
#           'FiftyPercentChance'       => sub { return $qwenv::random < 0.50 },
#           'SeventyFivePercentChance' => sub { return $qwenv::random < 0.75 }
#       );
#       my %needArgs = (
#           'CameFrom'      => 1,
#           'NotComingFrom' => 1
#       );
#   
#       my $_reportError = sub {
#           my ($err) = @_;
#           $$errBuf .= "<tt><b><u>$err near:</u></b></tt><br>\n$line";
#           $out .= qq{<font color="FF7700">Story Markup Problem</font><br>\n}
#                 . qq{"<tt><b><u>$err near:</u></b></tt><br>\n`$line\n\n"};
#       };
#   
#       #
#       # to check for:
#       #   multiple otherwise in block
#       #   25 < 50 < 75 order
#   
#       while ($text =~ /.*?(\n|$)/gs) {
#           $line = $&;
#           my $error = "";
#   
#   print "ISM: top inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";
#   
#           if ($line =~ /^``([A-Za-z]+)?\s*(.*?)\s*$/s) {
#               my ($op, $param) = ($1, $2);
#               $priorCaseSucceeded = $thisCaseSucceeded;
#               $thisCaseSucceeded = 0;
#   print "ISM: isblk inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";
#               unless ($op || $param) {
#   print "ISM: end_block\n<br>";
#                   $inBlock = 0;
#               }
#               elsif ($op) {
#                   $inBlock = 1;
#                   $param = "" unless defined($param);
#   print "ISM: iscmd op($op) param($param) inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";
#                   my @args = split(/[^A-Za-z]/, $param);
#                   if ($needArgs{$op}  &&  ! scalar(@args)) {
#                       &$_reportError("One or more PageNames needed for $op");
#                   }
#                   else {
#                       unless ($priorCaseSucceeded) {
#                           my $func = $opExe{$op};
#                           if (defined($func)) {
#                               $thisCaseSucceeded = &$func(@args);
#                           }
#                           else {
#                               &$_reportError("Unrecognized Story Markup Command $op");
#                           }
#   print "ISM: func op($op) param($param) inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";
#                       }
#                   }
#               }
#           }
#           else {
#   print "ISM: out inB($inBlock) prior($priorCaseSucceeded) this($thisCaseSucceeded)\n<br>";
#               $out .= $line unless ($inBlock && ! $thisCaseSucceeded);
#           }
#       }
#   
#       return $out;
#   }






sub markupWikiLinks {
    my ($pageTextRef) = @_;

    {
        local($^W) = 0;  # warnings off temporarily - the undef values are intentionally used below
        # \x60 below is back-tick (`)
        my $sub;
        $$pageTextRef =~ s{\b(mailto|http|ftp|wiki):}{lc($&)}gsei;
        $$pageTextRef =~ s` (?: \b((?:[A-Z][a-z]*){2,})\b ) |
                            ( \b mailto:[^\s]+?[@][^\s)'\x60<]+ ) |
                            ( \b http:[^\s)'\x60<]+ ) |
                            ( \b ftp://[^\s)'\x60<]+ ) |
                            ( \b wiki:[A-Za-z]+ )
                          ` ($1)? wikiLink($1): markupUrl("$2$3$4$5");
                          `gsex;
    }
}

sub stripWikiMetaCharsInHtml {
    my ($textRef) = @_;
    
    {
        local($^W) = 0;  # warnings off temporarily - the undef values are intentionally used below
        my ($text, $tag);
        # Don't modify content between html < tags >.
        $$textRef =~ s{ ([^<]*) (<[^>]*> | $ ) }
                      { ($text, $tag) = ($1, $2);
                        stripWikiMetaChars(\$text);
                        "$text$tag";
                      }gsex;
    }
}

sub stripWikiMetaChars {
    my ($textRef) = @_;
    $$textRef =~ s/`(`*)/$1/gse;  # gobble one back-tick (WikiName`s handler)
}

sub markupUrl {
    my ($url) = @_;
    if ($url =~ m{^http:.*\.(gif|jpg|jpeg|png)$}i) {
      # $url =~ s{^http:([^/])}{$1};  # strip leading http: for relative urls
        $url =~ s{^http:([^/])}{$qwcfg::serverUrl/$1};
        return img({src => $url});
    }
    elsif ($url =~ m{^wiki:(.*)$}i) {
        my $pageName = $1;
        my $url = "$qwcfg::interWikiUrl$pageName";
        return qq{<a href="$url" target="_blank">}
             . qq{<font size="-3">Wiki:</font>$pageName</a>};
    }
    else {
        my $dispUrl = $url;
      # $url =~ s{^http:([^/])}{$1};  # strip leading http: for relative urls
        $url =~ s{^http:([^/])}{$qwcfg::serverUrl/$1};
        return a({-href => $url, -target => "_top"}, $dispUrl);  # break out of frames
    }
}

sub wikiLink {
    my ($name) = @_;
    my $strippedName = $name;
    stripWikiMetaChars(\$strippedName);
    if (exists($qwdbm::pages{$strippedName})) {
        # If name exists in database, mark it up (even if non-strict WikiName)
        return a({href=>"$qwcfg::qwikiUrl?$strippedName"}, $name);
    }
    else {
        # Only automatically provide edit links for strict WikiNames.
        # Names like CLanguage or BASIC must be created manually (via
        # the edit URL), but will then be marked up once they exist
        # in the database.
        return (validStrictWikiName($strippedName))?
            ($name . a({href=>"$qwcfg::qwikiUrl?edit=$strippedName"}, "?")) :
            ($name);
    }
}

sub validWikiName {
    my ($name) = @_;
    return $name =~ /^(?:[A-Z][a-z]*){2,}$/s;  # (allows CLanguage, BASIC, etc.)
}

sub validStrictWikiName {
    my ($name) = @_;
    return $name =~ /^(?:[A-Z][a-z]+){2,}$/s;  # (strict AbCd names)
}

sub indexWikiPage {
    my ($pageName, $pageText) = @_;
    # Break up inter-wiki names so not indexed locally?  [maybe no need...]
    # $pageText =~ s/\bwiki:([a-z]+)/separateWikiWord($1)/gsei;
    stripWikiMetaChars(\$pageText);
    my $sepWords = join(" ", splitWikiWords($pageText));
    $qwdbm::search{$pageName} = $pageText . " " . $sepWords;
}

# Somewhat of a kludge for the fulltext search.  Since it's a literal
# keyword search, we're helping it out by breaking up WikiNames into
# their component words.  (The full WikiNames are also indexed, elsewhere.)
# -given
#   "NormalWikiWord WordIMARSEmbedded PLAINCAPS CLanguage BlaZEE DEDBoeCON"
# -returns (unordered)
#   CON ZEE Language Wiki Boe Normal Embedded PLAINCAPS Word IMARS DED Bla
#
sub splitWikiWords {
    my ($pageText) = @_;
    my %words;
    while ($pageText =~ /\b((?:[A-Z][a-z]*){2,})\b/gs) {
        my $wikiWord = $&;
        while ($wikiWord =~ /[A-Z](?:[a-z]+|(?:[A-Z](?![a-z]))+)/g) {
            $words{"$&"} = 1;
        }
    }
    return keys(%words);
}

sub separateWikiWord {
    my ($wikiWord) = @_;
    my $sep = "";
    while ($wikiWord =~ /[A-Z](?:[a-z]+|(?:[A-Z](?![a-z]))+)/g) {
        $sep .= " " if ($sep);
        $sep .= $&;
    }
    return $sep;
}

sub getEditDate {
    my ($saveTime) = @_;
    return strftime("%b %d %Y", localtime($saveTime));
}

sub getDiffTime {
    local($^W) = undef;
    my ($old, $new) = @_;

    my $sec = abs($new - $old);
    my $min = int($sec / 60) % 60;
    my $hour = int($sec / 3600) % 24;
    my $day = int($sec / 86400) % 7;
    my $week = int($sec / (86400*7)) % 52;
    my $year = int($sec / (86400*364));  # (52 weeks exactly... close enough :)
    $min = 1 if ($sec < 60);
    
    my @comp;
    push(@comp, "$year year"  . (($year == 1)? "" : "s")) if ($year);
    push(@comp, "$week week"  . (($week == 1)? "" : "s")) if ($week);
    push(@comp, "$day day"    . (($day == 1)?  "" : "s")) if ($day);
    push(@comp, "$hour hour"  . (($hour == 1)? "" : "s")) if ($hour);
    push(@comp, "$min minute" . (($min == 1)?  "" : "s"));
    return join(', ', @comp);
}

sub timeSinceRevEdit {
    my ($rev, $revTime, $curTime, $authAddr) = @_;

    my $diffTime = getDiffTime($revTime, $curTime);
    return qq{<tt><b><u>}
         . qq{Revision $rev was edited $diffTime ago by $authAddr.}
         . qq{</u></b></tt>\n};
}

sub getHostName {
    my ($hostAddr) = @_;
    return "unknown" unless ($hostAddr  &&  $hostAddr ne "unknown");
    my $iaddr = inet_aton($hostAddr);
    return "$hostAddr-unknown" unless defined($iaddr);
    my $name = gethostbyaddr($iaddr, AF_INET);
    return ($name)? $name : $hostAddr;
}

sub simpleHashFromString {
    local($^W) = 0;  # warnings off temporarily - the undef values are intentionally used below
    my ($str) = @_;

    my $hash = {};
    while ($str =~ /([^\012]*)\012([^\012]*)\012/gs) {
        my ($key, $val) = ($1, $2);
        _decodeHashStringElem(\$key);
        _decodeHashStringElem(\$val);
        $$hash{$key} = $val;
    }
    return $hash;
}

sub _decodeHashStringElem {
    my ($elemRef) = @_;
    if ($$elemRef eq "\011") {
        $$elemRef = undef;
    }
    else {
        $$elemRef =~ s/[%]0a/\012/gsi;
        $$elemRef =~ s/[%]09/\011/gs;
        $$elemRef =~ s/[%]25/%/gs;
    }
}

sub simpleHashToString {
    my ($hash) = @_;

    my $str = "";
    my $elem;
    foreach $elem (%$hash) {
        if (defined($elem)) {
            $elem =~ s/[%]/%25/gs;
            $elem =~ s/[\011]/%09/gs;  # tab-char used for undef
            $elem =~ s/[\012]/%0a/gs;  # newline used as field separator below
        }
        else {
            $elem = "\011";
        }
        $str .= "$elem\012";
    }
    return $str;
}


1;

# eof

